# Written by John Hoffman
# see LICENSE.txt for license information
# Updated and modified for ABC_OKC : Old King Cole

import socket
from httplib import HTTPConnection, HTTPSConnection, HTTPException
from urlparse import urlparse
from gzip import GzipFile
from StringIO import StringIO
from random import getrandbits
from binascii import b2a_hex
from bencode import bdecode
from utility import exceptionArgsToString, getABCUtility

from __init__ import USERAGENT

MAX_REDIRECTS = 10

def toint(s):
    return long(b2a_hex(s), 16)

def tobinary(i):
    return (chr(i >> 24) + chr((i >> 16) & 0xFF) + 
            chr((i >> 8) & 0xFF) + chr(i & 0xFF))


class urlopen:
    def __init__(self, url):
        self.tries = 0
        self.gzipfile = None
        self._open(url.strip())

    def _open(self, url):
        self.tries += 1
        if self.tries > MAX_REDIRECTS:
            raise IOError, (getABCUtility().lang.get('httperror'), 500,
                            getABCUtility().lang.get('redirrecursion'))
        scheme, netloc, path, pars, query, fragment = urlparse(url)
        if scheme != 'http' and scheme != 'https':
            raise IOError, (getABCUtility().lang.get('urlerror'), getABCUtility().lang.get('unknownurltype'), scheme, url)
        url = path
        if pars:
            url += ';' + pars
        if query:
            url += '?' + query
        try:
            if scheme == 'http':
                self.connection = HTTPConnection(netloc, timeout = 30)
            else:
                self.connection = HTTPSConnection(netloc, timeout = 30)
            self.connection.request('GET', url, None,
                                    {'User-Agent': USERAGENT[0],
                                     'Accept-Encoding': 'gzip'})
            self.response = self.connection.getresponse()
        except HTTPException, e:
            raise IOError, (getABCUtility().lang.get('httperror'), exceptionArgsToString(e))
        status = self.response.status
        if status == 301 or status == 302:
            try:
                self.connection.close()
            except:
                pass
            self._open(self.response.getheader('Location'))
        elif status != 200:
            raise IOError, (getABCUtility().lang.get('httperror'), status, self.response.reason)

    def read(self):
        try:
            data = self.response.read()
        except:
            raise IOError, (getABCUtility().lang.get('httperror'), getABCUtility().lang.get('corruptresponse'))
        if self.response.getheader('Content-Encoding', '').find('gzip') >= 0:
            try:
                self.gzipfile = GzipFile(fileobj = StringIO(data))
                data = self.gzipfile.read()
            except:
                raise IOError, (getABCUtility().lang.get('httperror'), getABCUtility().lang.get('corruptresponse'))
        return data

    def close(self):
        self.connection.close()
        if self.gzipfile:
            self.gzipfile.close()


def udpurlopen(url, request):
    # Max timeout for each request is 30 s
    up = urlparse(url.strip())
    if up.port is None:
        raise IOError, (getABCUtility().lang.get('urlerror'), getABCUtility().lang.get('noport'))
    ip = getABCUtility().getHostByName(up.hostname)
    if ip is None:
        raise IOError, (getABCUtility().lang.get('urlerror'), getABCUtility().lang.get('cantresolve') + up.hostname)

    # Get connection ID
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.settimeout(15)
    for reqnumber in xrange(2):
        transactionid = tobinary(getrandbits(32))
        sock.sendto("\x00\x00\x04\x17'\x10\x19\x80" + tobinary(0) + transactionid, 0, (ip, up.port))
        try:
            data = sock.recv(512)
        except socket.timeout:
            if reqnumber == 0:
                continue
            sock.close()
            raise IOError, (getABCUtility().lang.get('udperror'), getABCUtility().lang.get('timeout'))
        break
    if len(data) < 16:
        sock.close()
        raise IOError, (getABCUtility().lang.get('udperror'), getABCUtility().lang.get('corruptresponse'))
    receivedaction = toint(data[0:4])
    receivedtransactionid = data[4:8]
    if receivedtransactionid == transactionid and receivedaction == 3:
        sock.close()
        raise IOError, (getABCUtility().lang.get('udperror'), data[8:])
    if receivedtransactionid != transactionid or receivedaction != 0:
        sock.close()
        raise IOError, (getABCUtility().lang.get('udperror'), getABCUtility().lang.get('corruptresponse'))
    connectionid = data[8:16]

    # Send request
    for reqnumber in xrange(2):
        transactionid = tobinary(getrandbits(32))
        sock.sendto(connectionid + tobinary(request[0]) + transactionid + request[1], 0, (ip, up.port))
        try:
            data = sock.recv(4096)
        except socket.timeout:
            if reqnumber == 0:
                continue
            sock.close()
            raise IOError, (getABCUtility().lang.get('udperror'), getABCUtility().lang.get('timeout'))
        break
    sock.close()
    receivedaction = toint(data[0:4])
    receivedtransactionid = data[4:8]
    if receivedtransactionid == transactionid and receivedaction == 3:
        raise IOError, (getABCUtility().lang.get('udperror'), data[8:])
    if receivedtransactionid != transactionid or receivedaction != request[0]:
        raise IOError, (getABCUtility().lang.get('udperror'), getABCUtility().lang.get('corruptresponse'))

    return data
